import { PaymentTx_T, ParameterizedFunction, Payload, RJSVM, RJSVM_Config, RJSVM_Implementations, Function_Map, Endpoints_Of, State_Of, Generic_Ctor_ReturnType, RJSVM_Endpoint } from "./types"
import { Client as Xrpl } from 'xrpl';
import { xrpIO } from 'xrpio';
import { DataParser } from "../../util/dataparser";
import { XRP_ADDRESS } from "../../util/protocol.constants";
import { payloadSchema } from "./schemas";
import { InsufficientFeeError, RestrictedAccessError } from "../../util/errors";


export abstract class RJSVM_Builder {
    static from<
        Base_Ctor extends (new (...args: any) => RJSVM<State_T, Endpoints_T>) | (abstract new (...args: any) => RJSVM<State_T, Endpoints_T>),
        Impl extends RJSVM_Implementations<any, Endpoints_T>,
        Endpoints_T extends Function_Map = Endpoints_Of<Generic_Ctor_ReturnType<Base_Ctor>>,
        State_T = State_Of<Generic_Ctor_ReturnType<Base_Ctor>>
    >(Base: Base_Ctor, defs: Impl): (new (config: RJSVM_Config) => RJSVM<State_T, Endpoints_T>) {
        return <any>class RJSVM_Runnable extends (Base as any) {

            sync_block_height = -1

            private rippleApi: Xrpl
            private xrpIO: xrpIO
            private syncTimeout: NodeJS.Timeout
            private subscribers = {}
            private onceSubscribers = {}

            public readonly definitions: Impl

            constructor(
                public readonly config: RJSVM_Config
            ) {
                super()

                if (!XRP_ADDRESS.test(this.owner)) {
                    const err = new Error(`Inavlid owner address ${this.owner}`)
                    this.emit('error', err)
                    throw err
                }

                this.definitions = Object.freeze(defs)
                Object.entries(this.definitions).forEach(([k, v]) => {
                    this[k] = (env: PaymentTx_T, ...args: any) => (v.implementation as ParameterizedFunction).apply(this as any, [env, ...args])
                })
            }

            public connect = async () => {
                this.rippleApi = new Xrpl(this.config.rippleNode)
                await this.rippleApi.connect()

                this.xrpIO = new xrpIO(this.config.rippleNode, { readFreshApi: false });
                await this.xrpIO.connect()
                await this.sync()

                /*
                prototype for a new-block listener. would call on every closed ledger (i.e. new block)
                could be used to implement limited lifetime states without requiring new events to trigger

                this.rippleApi.on('ledgerClosed', (ledger: any) => {
                    console.log("---- Ledger ----")
                    console.log("index",ledger.ledger_index)
                    console.log("hash", ledger.ledger_hash)
                    console.log("---- /Ledger ----")
                })

                await this.rippleApi.request({
                    command: 'subscribe',
                    streams: ['ledger']
                })
                */

            }

            public disconnect = async () => {
                if (this.syncTimeout) {
                    clearTimeout(this.syncTimeout)
                    this.syncTimeout = undefined
                }
                if (this.xrpIO) {
                    await this.xrpIO.disconnect()
                    this.xrpIO = undefined
                }
                if (this.rippleApi) {
                    await this.rippleApi.disconnect()
                    this.rippleApi = undefined
                }

                this.subscribers = {}
                this.onceSubscribers = {}
            }

            public on = (event: string, handler: ParameterizedFunction) => {
                const availableEvents = ['error', ...Object.keys(this.definitions)]
                if (!availableEvents.includes(event))
                    return

                if (!this.subscribers[event])
                    this.subscribers[event] = []

                this.subscribers[event].push(handler)
            }

            public once = (event: string, handler: ParameterizedFunction) => {
                const availableEvents = ['error', ...Object.keys(this.definitions)]
                if (!availableEvents.includes(event))
                    return

                if (!this.onceSubscribers[event])
                    this.onceSubscribers[event] = []

                this.onceSubscribers[event].push(handler)
            }

            private emit = (event: string, payload: any) => {
                if (this.subscribers[event])
                    this.subscribers[event].forEach(handler => handler(payload))

                if (this.onceSubscribers[event]) {
                    this.onceSubscribers[event].forEach(handler => handler(payload))
                    this.onceSubscribers[event] = []
                }
            }

            private handlePayload = async (tx: PaymentTx_T, payload: Payload) => {
                if (!(payload.endpoint in this.definitions)) {
                    return
                }

                const endpointDef: RJSVM_Endpoint<RJSVM, any> = this.definitions[payload.endpoint]

                if (endpointDef.visibility === 'owner' && tx.Account !== this.owner) {
                    const err = new RestrictedAccessError(payload.endpoint, tx.hash, tx.Account, this.owner)
                    this.emit('error', err)
                    return
                }

                if (endpointDef.fee && Number(tx.Amount) < endpointDef.fee) {
                    const err = new InsufficientFeeError(tx.hash, Number(endpointDef.fee), Number(tx.Amount))
                    this.emit('error', err)
                    return
                }

                try {
                    const data = await this.xrpIO.treeRead([payload.data], tx.Account)
                    const jsonData = JSON.parse(data)
                    endpointDef.parameterSchema.parse(jsonData)
                    this[payload.endpoint].apply(this, [tx, jsonData])
                    this.emit(payload.endpoint, jsonData)
                } catch (err) {
                    this.emit('error', err)
                    return
                }
            }

            private parseMemos = async (tx: PaymentTx_T) => {
                const memos = tx.Memos
                memos
                    .map((memo: any) => {
                        if (!memo.Memo || !memo.Memo.MemoData) {
                            return
                        }
                        try {
                            const data = DataParser.hex_to_ascii(memo.Memo.MemoData)
                            return JSON.parse(data)
                        } catch (err) {
                            this.emit('error', err)
                            return
                        }
                    })
                    .filter((data: string) => data != undefined)
                    .forEach(payload => {
                        try {
                            const parsedPayload = payloadSchema.parse(payload)
                            this.handlePayload(tx, parsedPayload)
                        } catch (err) {
                            this.emit('error', err)
                            return
                        }
                    })
            }

            private sync = async (marker?: any): Promise<void> => {
                //If `forward` is set to true, returns values indexed with the oldest ledger first. 
                //Otherwise, the results are indexed with the newest ledger first. 
                //(Each page of results may not be internally ordered, but the pages are overall ordered.)

                const resp = await this.rippleApi.request({
                    command: "account_tx",
                    account: this.config.listeningAddress,
                    forward: true,
                    ledger_index_min: this.sync_block_height,
                    marker: marker,
                })

                const full_txs = resp.result.transactions
                    .filter(tx => !!tx.tx_json.Memos)
                    .map(tx => {
                        return ({
                            ...tx.tx_json,
                            hash: tx.hash,
                            Amount: tx.tx_json["DeliverMax"],
                            inLedger: tx.tx_json['ledger_index']
                        } as PaymentTx_T)
                    })
                
                await Promise.all(
                    full_txs
                        .map(tx => this.parseMemos(tx))
                )

                //if marker is present the result is paginated.
                //re-run the same request with the marker included
                if (resp.result.marker) {
                    await this.sync(resp.result.marker)
                } else {
                    //presence of no marker means we caught up to current block height
                    this.sync_block_height = resp.result.ledger_index_max + 1
                    //schedule the next sync
                    this.syncTimeout = setTimeout(this.sync, 5000)
                }
            }
        }
    }
}